@@ -68,6 +68,12 @@
 #include "nsCSSFrameConstructor.h"
 #include "nsGUIEvent.h"
 #include "nsIRootBox.h"
+#include "nsIFocusController.h"
+#include "nsIDocShellTreeItem.h"
+#include "nsIDocShell.h"
+#include "nsPIDOMWindow.h"
+#include "nsIInterfaceRequestorUtils.h"
+#include "nsIBaseWindow.h"
 
 #define NS_MENU_POPUP_LIST_INDEX   0
 
@@ -342,8 +348,12 @@ nsPopupSetFrame::ShowPopup(nsIContent* aElementContent, nsIContent* aPopupConten
                            const nsString& aPopupType, const nsString& anAnchorAlignment,
                            const nsString& aPopupAlignment)
 {
+  if (!MayOpenPopup(this))
+    return NS_OK;
+
+  nsWeakFrame weakFrame(this);
   // First fire the popupshowing event.
-  if (!OnCreate(aXPos, aYPos, aPopupContent))
+  if (!OnCreate(aXPos, aYPos, aPopupContent) || !weakFrame.IsAlive())
     return NS_OK;
         
   // See if we already have an entry in our list.  We must create a new one on a miss.
@@ -377,19 +387,35 @@ nsPopupSetFrame::ShowPopup(nsIContent* aElementContent, nsIContent* aPopupConten
   // Generate the popup.
   entry->mCreateHandlerSucceeded = PR_TRUE;
   entry->mIsOpen = PR_TRUE;
+  // This may destroy or change entry->mPopupFrame or remove the entry from
+  // mPopupList. |this| may also get deleted.
   MarkAsGenerated(aPopupContent);
 
+  if (!weakFrame.IsAlive()) {
+    return NS_OK;
+  }
+
+  nsPopupFrameList* newEntry =
+    mPopupList ? mPopupList->GetEntry(aPopupContent) : nsnull;
+  if (!newEntry || newEntry != entry) {
+    NS_WARNING("The popup entry for aPopupContent has changed!");
+    return NS_OK;
+  }
+
   // determine if this menu is a context menu and flag it
-  nsIFrame* activeChild = entry->mPopupFrame;
   nsIMenuParent* childPopup = nsnull;
-  if (activeChild)
-    CallQueryInterface(activeChild, &childPopup);
+  if (entry->mPopupFrame)
+    CallQueryInterface(entry->mPopupFrame, &childPopup);
   if ( childPopup && aPopupType.EqualsLiteral("context") )
     childPopup->SetIsContextMenu(PR_TRUE);
 
   // Now open the popup.
   OpenPopup(entry, PR_TRUE);
 
+  if (!weakFrame.IsAlive()) {
+    return NS_OK;
+  }
+
   // Now fire the popupshown event.
   OnCreated(aXPos, aYPos, aPopupContent);
 
@@ -438,35 +464,39 @@ nsPopupSetFrame::DestroyPopup(nsIFrame* aPopup, PRBool aDestroyEntireChain)
   nsPopupFrameList* entry = mPopupList->GetEntryByFrame(aPopup);
 
   if (entry && entry->mCreateHandlerSucceeded) {    // ensure the popup was created before we try to destroy it
+    nsWeakFrame weakFrame(this);
     OpenPopup(entry, PR_FALSE);
-    entry->mPopupType.SetLength(0);
-  
-    if (aDestroyEntireChain && entry->mElementContent && entry->mPopupType.EqualsLiteral("context")) {
-      // If we are a context menu, and if we are attached to a
-      // menupopup, then destroying us should also dismiss the parent
-      // menu popup.
-      if (entry->mElementContent->Tag() == nsXULAtoms::menupopup) {
-        nsIFrame* popupFrame = nsnull;
-        mPresContext->PresShell()->GetPrimaryFrameFor(entry->mElementContent,
-                                                      &popupFrame);
-        if (popupFrame) {
-          nsIMenuParent *menuParent;
-          if (NS_SUCCEEDED(CallQueryInterface(popupFrame, &menuParent))) {
-            menuParent->DismissChain();
+    nsCOMPtr<nsIContent> popupContent = entry->mPopupContent;
+    if (weakFrame.IsAlive()) {
+      entry->mPopupType.SetLength(0);
+
+      if (aDestroyEntireChain && entry->mElementContent && entry->mPopupType.EqualsLiteral("context")) {
+        // If we are a context menu, and if we are attached to a
+        // menupopup, then destroying us should also dismiss the parent
+        // menu popup.
+        if (entry->mElementContent->Tag() == nsXULAtoms::menupopup) {
+          nsIFrame* popupFrame = nsnull;
+          mPresContext->PresShell()->GetPrimaryFrameFor(entry->mElementContent,
+                                                        &popupFrame);
+          if (popupFrame) {
+            nsIMenuParent *menuParent;
+            if (NS_SUCCEEDED(CallQueryInterface(popupFrame, &menuParent))) {
+              menuParent->DismissChain();
+            }
           }
         }
       }
+  
+      // clear things out for next time
+      entry->mCreateHandlerSucceeded = PR_FALSE;
+      entry->mElementContent = nsnull;
+      entry->mXPos = entry->mYPos = 0;
+      entry->mLastPref.width = -1;
+      entry->mLastPref.height = -1;
     }
 
-    // clear things out for next time
-    entry->mCreateHandlerSucceeded = PR_FALSE;
-    entry->mElementContent = nsnull;
-    entry->mXPos = entry->mYPos = 0;
-    entry->mLastPref.width = -1;
-    entry->mLastPref.height = -1;
-
     // ungenerate the popup.
-    entry->mPopupContent->UnsetAttr(kNameSpaceID_None, nsXULAtoms::menugenerated, PR_TRUE);
+    popupContent->UnsetAttr(kNameSpaceID_None, nsXULAtoms::menugenerated, PR_TRUE);
   }
 
   return NS_OK;
@@ -490,21 +520,28 @@ nsPopupSetFrame::MarkAsGenerated(nsIContent* aPopupContent)
 void
 nsPopupSetFrame::OpenPopup(nsPopupFrameList* aEntry, PRBool aActivateFlag)
 {
+  nsWeakFrame weakFrame(this);
+  nsIFrame* activeChild = aEntry->mPopupFrame;
+  nsWeakFrame weakPopupFrame(activeChild);
+  nsRefPtr<nsPresContext> presContext = GetPresContext();
+  nsCOMPtr<nsIContent> popupContent = aEntry->mPopupContent;
+  PRBool createHandlerSucceeded = aEntry->mCreateHandlerSucceeded;
+  nsAutoString popupType = aEntry->mPopupType;
   if (aActivateFlag) {
     ActivatePopup(aEntry, PR_TRUE);
 
     // register the rollup listeners, etc, but not if we're a tooltip
-    if (!aEntry->mPopupType.EqualsLiteral("tooltip")) {
+    if (!popupType.EqualsLiteral("tooltip")) {
       nsIFrame* activeChild = aEntry->mPopupFrame;
       nsIMenuParent* childPopup = nsnull;
-      if (activeChild)
+      if (weakPopupFrame.IsAlive())
         CallQueryInterface(activeChild, &childPopup);
 
       // Tooltips don't get keyboard navigation
       if (childPopup && !nsMenuFrame::sDismissalListener) {
         // First check and make sure this popup wants keyboard navigation
         nsAutoString property;    
-        aEntry->mPopupContent->GetAttr(kNameSpaceID_None, nsXULAtoms::ignorekeys, property);
+        popupContent->GetAttr(kNameSpaceID_None, nsXULAtoms::ignorekeys, property);
         if (!property.EqualsLiteral("true"))
           childPopup->InstallKeyboardNavigator();
       }
@@ -513,29 +550,32 @@ nsPopupSetFrame::OpenPopup(nsPopupFrameList* aEntry, PRBool aActivateFlag)
     }
   }
   else {
-    if (aEntry->mCreateHandlerSucceeded && !OnDestroy(aEntry->mPopupContent))
+    if (createHandlerSucceeded && !OnDestroy(popupContent))
       return;
 
     // Unregister, but not if we're a tooltip
-    if (!aEntry->mPopupType.EqualsLiteral("tooltip") ) {
+    if (!popupType.EqualsLiteral("tooltip")) {
       if (nsMenuFrame::sDismissalListener)
         nsMenuFrame::sDismissalListener->Unregister();
     }
     
     // Remove any keyboard navigators
     nsIMenuParent* childPopup = nsnull;
-    if (aEntry->mPopupFrame)
-      CallQueryInterface(aEntry->mPopupFrame, &childPopup);
+    if (weakPopupFrame.IsAlive())
+      CallQueryInterface(activeChild, &childPopup);
     if (childPopup)
       childPopup->RemoveKeyboardNavigator();
 
-    ActivatePopup(aEntry, PR_FALSE);
+    if (weakPopupFrame.IsAlive())
+      ActivatePopup(aEntry, PR_FALSE);
 
-    OnDestroyed(aEntry->mPopupContent);
+    OnDestroyed(presContext, popupContent);
   }
 
-  nsBoxLayoutState state(mPresContext);
-  MarkDirtyChildren(state); // Mark ourselves dirty.
+  if (weakFrame.IsAlive()) {
+    nsBoxLayoutState state(mPresContext);
+    MarkDirtyChildren(state); // Mark ourselves dirty.
+  }
 }
 
 void
@@ -549,14 +589,17 @@ nsPopupSetFrame::ActivatePopup(nsPopupFrameList* aEntry, PRBool aActivateFlag)
       // XXXben hook in |width| and |height| usage here? 
       aEntry->mPopupContent->SetAttr(kNameSpaceID_None, nsXULAtoms::menutobedisplayed, NS_LITERAL_STRING("true"), PR_TRUE);
     else {
-      aEntry->mPopupContent->UnsetAttr(kNameSpaceID_None, nsXULAtoms::menuactive, PR_TRUE);
-      aEntry->mPopupContent->UnsetAttr(kNameSpaceID_None, nsXULAtoms::menutobedisplayed, PR_TRUE);
+      nsWeakFrame weakFrame(this);
+      nsWeakFrame weakActiveChild(aEntry->mPopupFrame);
+      nsCOMPtr<nsIContent> content = aEntry->mPopupContent;
+      content->UnsetAttr(kNameSpaceID_None, nsXULAtoms::menuactive, PR_TRUE);
+      content->UnsetAttr(kNameSpaceID_None, nsXULAtoms::menutobedisplayed, PR_TRUE);
 
       // get rid of the reflows we just created. If we leave them hanging around, we
       // can get into trouble if a dialog with a modal event loop comes along and
       // processes the reflows before we get to call DestroyChain(). Processing the
       // reflow will cause the popup to show itself again. (bug 71219)
-      nsIDocument* doc = aEntry->mPopupContent->GetDocument();
+      nsIDocument* doc = content->GetDocument();
       if (doc)
         doc->FlushPendingNotifications(Flush_OnlyReflow);
          
@@ -564,8 +607,8 @@ nsPopupSetFrame::ActivatePopup(nsPopupFrameList* aEntry, PRBool aActivateFlag)
       // since we could be cleaning up after someone that didn't correctly 
       // destroy the popup.
       nsIFrame* activeChild = aEntry->mPopupFrame;
-      if (activeChild) {
-        nsIView* view = activeChild->GetView();
+      if (weakFrame.IsAlive() && weakActiveChild.IsAlive()) {
+        nsIView* view = weakActiveChild.GetFrame()->GetView();
         NS_ASSERTION(view, "View is gone, looks like someone forgot to roll up the popup!");
         if (view) {
           nsIViewManager* viewManager = view->GetViewManager();
@@ -574,7 +617,7 @@ nsPopupSetFrame::ActivatePopup(nsPopupFrameList* aEntry, PRBool aActivateFlag)
           viewManager->ResizeView(view, r);
           if (aEntry->mIsOpen) {
             aEntry->mIsOpen = PR_FALSE;
-            FireChromeDOMEvent(NS_LITERAL_STRING("DOMMenuInactive"), aEntry->mPopupContent);
+            FireChromeDOMEvent(NS_LITERAL_STRING("DOMMenuInactive"), content);
           }
         }
       }
@@ -592,11 +635,11 @@ nsPopupSetFrame::OnCreate(PRInt32 aX, PRInt32 aY, nsIContent* aPopupContent)
   event.point.y = aY;
 
   if (aPopupContent) {
-    nsIPresShell *shell = mPresContext->GetPresShell();
+    nsCOMPtr<nsIContent> kungFuDeathGrip(aPopupContent);
+    nsCOMPtr<nsIPresShell> shell = mPresContext->GetPresShell();
     if (shell) {
       nsresult rv = shell->HandleDOMEventWithTarget(aPopupContent, &event,
                                                     &status);
-      // shell may no longer be alive, don't use it here unless you keep a ref
       if ( NS_FAILED(rv) || status == nsEventStatus_eConsumeNoDefault )
         return PR_FALSE;
     }
@@ -610,7 +653,7 @@ nsPopupSetFrame::OnCreate(PRInt32 aX, PRInt32 aY, nsIContent* aPopupContent)
 
     PRUint32 count = aPopupContent->GetChildCount();
     for (PRUint32 i = 0; i < count; i++) {
-      nsIContent *grandChild = aPopupContent->GetChildAt(i);
+      nsCOMPtr<nsIContent> grandChild = aPopupContent->GetChildAt(i);
 
       if (grandChild->Tag() == nsXULAtoms::menuitem) {
         // See if we have a command attribute.
@@ -670,11 +713,10 @@ nsPopupSetFrame::OnCreated(PRInt32 aX, PRInt32 aY, nsIContent* aPopupContent)
   event.point.y = aY;
 
   if (aPopupContent) {
-    nsIPresShell *shell = mPresContext->GetPresShell();
+    nsCOMPtr<nsIPresShell> shell = mPresContext->GetPresShell();
     if (shell) {
       nsresult rv = shell->HandleDOMEventWithTarget(aPopupContent, &event,
                                                     &status);
-      // shell may no longer be alive, don't use it here unless you keep a ref
       if ( NS_FAILED(rv) || status == nsEventStatus_eConsumeNoDefault )
         return PR_FALSE;
     }
@@ -691,11 +733,10 @@ nsPopupSetFrame::OnDestroy(nsIContent* aPopupContent)
                      nsMouseEvent::eReal);
 
   if (aPopupContent) {
-    nsIPresShell *shell = mPresContext->GetPresShell();
+    nsCOMPtr<nsIPresShell> shell = mPresContext->GetPresShell();
     if (shell) {
       nsresult rv = shell->HandleDOMEventWithTarget(aPopupContent, &event,
                                                     &status);
-      // shell may no longer be alive, don't use it here unless you keep a ref
       if ( NS_FAILED(rv) || status == nsEventStatus_eConsumeNoDefault )
         return PR_FALSE;
     }
@@ -704,18 +745,18 @@ nsPopupSetFrame::OnDestroy(nsIContent* aPopupContent)
 }
 
 PRBool
-nsPopupSetFrame::OnDestroyed(nsIContent* aPopupContent)
+nsPopupSetFrame::OnDestroyed(nsPresContext* aPresContext,
+                             nsIContent* aPopupContent)
 {
   nsEventStatus status = nsEventStatus_eIgnore;
   nsMouseEvent event(PR_TRUE, NS_XUL_POPUP_HIDDEN, nsnull,
                      nsMouseEvent::eReal);
 
-  if (aPopupContent) {
-    nsIPresShell *shell = mPresContext->GetPresShell();
+  if (aPopupContent && aPresContext) {
+    nsCOMPtr<nsIPresShell> shell = aPresContext->GetPresShell();
     if (shell) {
       nsresult rv = shell->HandleDOMEventWithTarget(aPopupContent, &event,
                                                     &status);
-      // shell may no longer be alive, don't use it here unless you keep a ref
       if ( NS_FAILED(rv) || status == nsEventStatus_eConsumeNoDefault )
         return PR_FALSE;
     }
@@ -797,3 +838,42 @@ nsPopupSetFrame::AddPopupFrame(nsIFrame* aPopup)
   return NS_OK;
 }
 
+//static
+PRBool
+nsPopupSetFrame::MayOpenPopup(nsIFrame* aFrame)
+{
+  nsCOMPtr<nsISupports> cont = aFrame->GetPresContext()->GetContainer();
+  nsCOMPtr<nsIDocShellTreeItem> dsti = do_QueryInterface(cont);
+  if (!dsti)
+    return PR_FALSE;
+
+  // chrome shells can always open popups
+  PRInt32 type = -1;
+  if (NS_SUCCEEDED(dsti->GetItemType(&type)) && type == nsIDocShellTreeItem::typeChrome)
+    return PR_TRUE;
+
+  nsCOMPtr<nsIDocShell> shell = do_QueryInterface(dsti);
+  if (!shell)
+    return PR_FALSE;
+
+  nsCOMPtr<nsPIDOMWindow> win = do_GetInterface(shell);
+  if (!win)
+    return PR_FALSE;
+
+  // only allow popups in active windows
+  PRBool active;
+  nsIFocusController* focusController = win->GetRootFocusController();
+  focusController->GetActive(&active);
+  if (!active)
+    return PR_FALSE;
+
+  nsCOMPtr<nsIBaseWindow> baseWin = do_QueryInterface(shell);
+  if (!baseWin)
+    return PR_FALSE;
+
+  // only allow popups in visible frames
+  PRBool visible;
+  baseWin->GetVisibility(&visible);
+  return visible;
+}
+
